// Copyright 2014 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build ignore
// +build ignore

#include "textflag.h"

// See memmove Go doc for important implementation constraints.

// Register map
//
// dstin  R0
// src    R1
// count  R2
// dst    R3 (same as R0, but gets modified in unaligned cases)
// srcend R4
// dstend R5
// data   R6-R17
// tmp1   R14

// Copies are split into 3 main cases: small copies of up to 32 bytes, medium
// copies of up to 128 bytes, and large copies. The overhead of the overlap
// check is negligible since it is only required for large copies.
//
// Large copies use a software pipelined loop processing 64 bytes per iteration.
// The destination pointer is 16-byte aligned to minimize unaligned accesses.
// The loop tail is handled by always copying 64 bytes from the end.

// func memmove(to, from unsafe.Pointer, n uintptr)
TEXT runtime·memmove(SB), NOSPLIT|NOFRAME, $0-24
	LDL	R0, to+0(FP)
	LDL	R1, from+8(FP)
	LDL	R2, n+16(FP)
	BEQ	R2, copy0

	// Small copies: 1..16 bytes
	CMPLE	R2, $16, R28
	BNE R28,	copy16

	// Large copies
	CMPLE	R2, $128, R28
	BEQ R28, copy_long
	CMPLE	R2, $32, R28
	BEQ	R28, copy32_128

	// Small copies: 17..32 bytes.
	LDL R6, 0(R1)
	LDL R7, 8(R1)
	ADDL R1, R2, R4
	LDL R12, -8(R4)
	LDL R13, -16(R4)

	STL R6, (R0)
	STL R7, 8(R0)
	ADDL R0, R2, R5
	STL R12, -8(R5)
	STL R13, -16(R5)
	RET

// Small copies: 1..16 bytes.
copy16:
	ADDL	R1, R2, R4 // R4 points just past the last source byte
	ADDL	R0, R2, R5 // R5 points just past the last destination byte
	CMPLT	R2, $8, R28
	BNE R28,	copy7
	LDL	R6, (R1)
	LDL	R7, -8(R4)
	STL	R6, (R0)
	STL	R7, -8(R5)
	RET

copy7:
	CMPLT	R2, $4, R28
	BNE R28, copy3
	LDW	R6, (R1)
	LDW	R7, -4(R4)
	STW	R6, (R0)
	STW	R7, -4(R5)
	RET

copy3:
	CMPLT	R2, $2, R28
	BNE R28, copy1
	LDHU	R6, (R1)
	LDHU	R7, -2(R4)
	STH	R6, (R0)
	STH	R7, -2(R5)
	RET

copy1:
	LDBU	R6, (R1)
	STB	R6, (R0)

copy0:
	RET

	// Medium copies: 33..128 bytes.
copy32_128:
	ADDL	R1, R2, R4          // R4 points just past the last source byte
	ADDL	R0, R2, R5          // R5 points just past the last destination byte
	VLDD V7, (R1)
	VLDD V8, -32(R4)
	CMPLE R2, $64, R28
	BEQ R28, copy128
	VSTD V7, (R0)
	VSTD V8, -32(R5)
	RET

	// Copy 65..128 bytes.
copy128:
	VLDD V3, 32(R1)
	CMPLE R2, $96, R28
	BNE R28, copy96
	VLDD V1, -64(R4)
	VSTD V1, -64(R5)

copy96:
	VSTD V7, (R0)
	VSTD V3, 32(R0)
	VSTD V8, -32(R5)
	RET

	// Copy more than 128 bytes.
copy_long:
	ADDL R1, R2, R4 // R4 points just past the last source byte
	ADDL R0, R2, R5 // R5 points just past the last destination byte
	LDI	R7, R31
	LDI	R8, R31

	LDI R28, $1024
	CMPLT R2, R28, R28
	BNE R28,backward_check
	// feature detect to decide how to align
	LDI	R7, R0
	LDI R8, R5
	BR	backward_check
	// R7 and R8 are used here for the realignment calculation. In
	// the use_aligned_loads case, R7 is the src pointer and R8 is
	// srcend pointer, which is used in the backward copy case.
	// When doing aligned stores, R7 is the dst pointer and R8 is
	// the dstend pointer.

backward_check:
	// Use backward copy if there is an overlap.
	SUBL	R0, R1, R14
	BEQ	R14, copy0
	CMPULT	R2, R14, R28
	BEQ	R28, copy_long_backward

	// Copy 16 bytes and then align src (R1) or dst (R0) to 16-byte alignment.
	VLDD V1, (R1)    // Load A
	AND R7, $31, R14
	SUBL R1, R14, R1
	SUBL R0, R14, R3
	ADDL R2, R14, R2
	VLDD V2, 32(R1) // Load   B
	VSTD V1, (R0)   // Store A
	VLDD V1, 64(R1)   // Load     C
	ADDL R1, $64, R1

	// 96 bytes have been loaded; if less than 96+64 bytes remain, copy from the end
	SUBL R2, $160, R2
	BLE R2, copy64_from_end

loop64:
	VSTD V2, 32(R3)   // Store B
	VLDD V2, 32(R1)   // Load  B (next)
	VSTD V1, 64(R3)   // Store  C
	VLDD V1, 64(R1)   // Load   C (next)
	ADDL R3, $64, R3
	ADDL R1, $64, R1
	SUBL R2, $64, R2
	BGT R2, loop64

	// Write the last iteration and copy 64 bytes from the end.
copy64_from_end:
	VLDD V6, -64(R4) // Load     D
	VSTD V2, 32(R3)  // Store B
	VLDD V5, -32(R4) // Load       F
	VSTD V1, 64(R3) // Store   C
	VSTD V6, -64(R5)
	VSTD V5, -32(R5)
	RET

	// Large backward copy for overlapping copies.
	// Copy 16 bytes and then align srcend (R4) or dstend (R5) to 16-byte alignment.
copy_long_backward:
	VLDD V3, -32(R4) // Load A
	AND R8, $31, R14
	SUBL R4, R14, R4
	SUBL R2, R14, R2
	VLDD V2, -32(R4) // Load   B
	VSTD V3, -32(R5) // Store A
	VLDD V1, -64(R4) // Load     C
	SUBL R4, $64, R4
	SUBL R5, R14, R5
	SUBL R2, $128, R2
	BLT R2, copy64_from_start

loop64_backward:
	VSTD V2, -32(R5) // Store B
	VLDD V2, -32(R4) // Load B(next)
	VSTD V1, -64(R5) // Store C
	VLDD V1, -64(R4) // Load C(next)
	SUBL R5, $64, R5
	SUBL R4, $64, R4
	SUBL R2, $64, R2
	BGT R2, loop64_backward

	// Write the last iteration and copy 64 bytes from the start.
copy64_from_start:
	VLDD V5, 32(R1) // Load    D
	VSTD V2, -32(R5) // Store B
	VLDD V4, (R1)   // Load      F
	VSTD V1, -64(R5) // Store C
	VSTD V5, 32(R0) // Store D
	VSTD V4, (R0) // Store F
	RET

